from typing import Any

from .components import CrudComponent, MetablockEntity
from .extensions import Extensions, Plugins
from .spaces import Spaces


class OrgSpaces(Spaces):
    def list_create_url(self) -> str:
        return "%s/%s" % (self.root.url, self.name)


class OrgExtensions(Extensions):
    def list_create_url(self) -> str:
        return "%s/%s" % (self.root.url, self.name)


class OrgPlugins(Plugins):
    def list_create_url(self) -> str:
        return "%s/%s" % (self.root.url, self.name)


class Org(MetablockEntity):
    """Object representing an Organization"""

    @property
    def name(self) -> str:
        return self.data.get("short_name", "")

    @property
    def spaces(self) -> OrgSpaces:
        return OrgSpaces(self, "spaces")

    @property
    def plugins(self) -> OrgPlugins:
        return OrgPlugins(self, "plugins")

    @property
    def extensions(self) -> OrgExtensions:
        return OrgExtensions(self, "extensions")

    @property
    def members(self) -> "Members":
        return Members(self)

    @property
    def roles(self) -> "Roles":
        return Roles(self)

    async def add_info(self, **data: Any) -> dict:
        return await self.patch(f"{self.url}/info", json=data, wrap=self.root.wrap)


class Member(MetablockEntity):
    """Object representing a organization member"""


class Role(MetablockEntity):
    """Object representing a organization role"""


class Members(CrudComponent):
    """Metablock organizations"""

    Entity = Member

    @property
    def url(self) -> str:
        return f"{self.root.url}/{self.name}"


class Roles(CrudComponent):
    """Metablock organizations"""

    Entity = Role

    @property
    def url(self) -> str:
        return f"{self.root.url}/{self.name}"


class Orgs(CrudComponent):
    """Metablock organizations"""

    Entity = Org



from __future__ import annotations

from typing import Any

from aiohttp.formdata import FormData

from .components import Component, CrudComponent, MetablockEntity


# Space
class Space(MetablockEntity):
    """Object representing a space"""

    @property
    def blocks(self) -> "SpaceBlocks":
        return SpaceBlocks(self, "services")

    @property
    def services(self) -> "SpaceBlocks":
        return self.blocks

    @property
    def extensions(self) -> "SpaceExtensions":
        return SpaceExtensions(self, "extensions")


class Spaces(CrudComponent):
    """Spaces"""

    Entity = Space


# Service
class Block(MetablockEntity):
    """Object representing a block in a space"""

    @property
    def plugins(self) -> BlockPlugins:
        return BlockPlugins(self, "plugins")

    @property
    def deployments(self) -> CrudComponent:
        return CrudComponent(self, "deployments")

    async def config(self, *, callback: Any = None) -> dict:
        return await self.get(f"{self.url}/config", callback=callback)

    async def certificate(self, *, callback: Any = None) -> dict:
        return await self.get(f"{self.url}/certificate", callback=callback)

    async def ship(
        self, name: str, bundle: str, env: str = "stage", *, callback: Any = None
    ) -> dict:
        data = FormData()
        data.add_field("name", name)
        data.add_field("bundle", open(bundle, "rb"), filename=bundle)
        data.add_field("env", env)
        return await self.post(f"{self.url}/deployments", data=data, callback=callback)

    async def add_route(self, *, callback: Any = None, **kwargs: Any) -> dict:
        """Add a new route to the block"""
        return await self.post(f"{self.url}/routes", json=kwargs, callback=callback)

    async def update_route(
        self, name: str, *, callback: Any = None, **kwargs: Any
    ) -> dict:
        """Update a route in the block"""
        return await self.patch(
            f"{self.url}/routes/{name}", json=kwargs, callback=callback
        )


# For backward compatibility
Service = Block


class Blocks(CrudComponent):
    """Services"""

    Entity = Block


class SpaceBlocks(Blocks):
    def list_create_url(self) -> str:
        return "%s/%s" % (self.root.url, self.name)


# SpaceExtension
class SpaceExtension(MetablockEntity):
    """Object representing an SpaceExtension"""


class SpaceExtensions(CrudComponent):
    Entity = SpaceExtension

    def list_create_url(self) -> str:
        return "%s/%s" % (self.root.url, self.name)

    def delete_url(self, id_name: str) -> str:
        return f"{self.parent_url}/{id_name}"


# ServicePlugin


class BlockPlugin(MetablockEntity):
    """Object representing an ServicePlugin"""


class BlockPlugins(CrudComponent):
    Entity = BlockPlugin

    def list_create_url(self) -> str:
        return "%s/%s" % (self.root.url, self.name)


# Domain


class Domains(Component):
    async def check(self, domain: str) -> str:
        return await self.cli.get(f"{self.url}/check/{domain}")



from typing import Any

from .components import MetablockEntity, MetablockResponseError
from .orgs import Org, Orgs


class User(MetablockEntity):
    """Object representing a Metablock user"""

    @property
    def url(self) -> str:
        return f"{self.root.url}/user"

    async def orgs(self, **kwargs: Any) -> list[Org]:
        """List user organizations"""
        orgs = Orgs(self.root)
        kwargs.setdefault("wrap", lambda dl: [Org(orgs, d) for d in dl])
        return await self.cli.get(f"{self.url}/orgs", **kwargs)

    async def get_permissions(self, **kwargs: Any) -> list[dict]:
        """List user permissions"""
        return await self.cli.get(f"{self.url}/permissions", **kwargs)

    async def tokens(self, **kwargs: Any) -> list[dict]:
        return await self.cli.get(f"{self.url}/tokens", **kwargs)

    async def create_token(self, **kwargs: Any) -> dict:
        return await self.cli.post(f"{self.url}/tokens", **kwargs)

    async def delete_token(self, token_id: str, **kwargs: Any) -> dict:
        return await self.cli.delete(f"{self.url}/tokens/{token_id}", **kwargs)

    async def check_password(self, password: str, **kwargs: Any) -> bool:
        try:
            return await self.root.post(
                f"{self.url}/password", json=dict(password=password), **kwargs
            )
        except MetablockResponseError as exc:
            if exc.status == 401:
                return False
            else:
                raise



from typing import Any

from multidict import MultiDict


def as_list(key: str, data: dict) -> dict:
    v = data.get(key)
    if isinstance(v, str):
        v = [v]
    elif not isinstance(v, list):
        v = []
    data[key] = v
    return data


def as_dict(data: Any, key: str = "data") -> dict:
    return {key: data} if not isinstance(data, dict) else data


def as_params(*, params: dict | None = None, **kwargs: Any) -> MultiDict:
    d = MultiDict(params if params is not None else {})
    d.update(kwargs)
    return d



from typing import Any

from .components import CrudComponent, MetablockEntity
from .spaces import BlockPlugin


# Extension
class Extension(MetablockEntity):
    """Object representing an Extension"""


class Extensions(CrudComponent):
    """Extensions"""

    Entity = Extension


# Plugin
class Plugin(MetablockEntity):
    """Object representing an Plugin"""

    async def blocks(self, **kwargs: Any) -> list[BlockPlugin]:
        return await self.get(f"{self.url}/services", **kwargs)


class Plugins(CrudComponent):
    """Plugins"""

    Entity = Plugin



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, AsyncIterator, Awaitable, Callable, cast

from aiohttp import ClientResponse
from yarl import URL

from .utils import as_dict, as_params

if TYPE_CHECKING:  # pragma: no cover
    from .client import Metablock


Callback = Callable[[ClientResponse], Awaitable[Any]]


class MetablockError(Exception):
    pass


class MetablockResponseError(MetablockError):
    def __init__(self, response: ClientResponse, message: Any = "") -> None:
        self.response = response
        self.message = as_dict(message, "message")
        self.message["request_url"] = str(response.url)
        self.message["request_method"] = response.method
        self.message["response_status"] = response.status

    @property
    def status(self) -> int:
        return self.response.status

    def __str__(self) -> str:
        return json.dumps(self.message, indent=4)


class HttpComponent(ABC):
    @abstractmethod
    def execute(self, url: str, *, method: str = "", **kwargs: Any) -> Any:
        """Execute Http request"""

    async def get(self, url: str, **kwargs: Any) -> Any:
        kwargs["method"] = "GET"
        return await self.execute(url, **kwargs)

    async def patch(self, url: str, **kwargs: Any) -> Any:
        kwargs["method"] = "PATCH"
        return await self.execute(url, **kwargs)

    async def post(self, url: str, **kwargs: Any) -> Any:
        kwargs["method"] = "POST"
        return await self.execute(url, **kwargs)

    async def put(self, url: str, **kwargs: Any) -> Any:
        kwargs["method"] = "PUT"
        return await self.execute(url, **kwargs)

    async def delete(self, url: str, **kwargs: Any) -> Any:
        kwargs["method"] = "DELETE"
        return await self.execute(url, **kwargs)

    def wrap(self, data: Any) -> Any:
        return data


class MetablockEntity(HttpComponent):
    """A Metablock entity"""

    __slots__ = ("root", "data")

    def __init__(
        self, root: Metablock | CrudComponent | MetablockEntity, data: dict
    ) -> None:
        self.root = root
        self.data = data

    def __repr__(self) -> str:
        return repr(self.data)

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(self, item: str) -> Any:
        return self.data[item]

    def __contains__(self, item: str) -> bool:
        return item in self.data

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.data == other.data

    def _asdict(self) -> dict:
        return self.data

    @property
    def cli(self) -> Metablock:
        return self.root.cli

    @property
    def id(self) -> str:
        return self.data.get("id", "")

    @property
    def name(self) -> str:
        return self.data.get("name", "")

    @property
    def url(self) -> str:
        return "%s/%s" % (self.root.url, self.id)

    async def execute(self, url: str | URL, **kwargs: Any) -> Any:
        return await self.cli.execute(url, **kwargs)

    def nice(self) -> str:
        return json.dumps(self.data, indent=4)


class Component:
    __slots__ = ("root", "name")

    def __init__(
        self, root: Metablock | CrudComponent | MetablockEntity, name: str = ""
    ) -> None:
        self.root = root
        self.name = name or self.__class__.__name__.lower()

    def __repr__(self) -> str:
        return self.url

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def cli(self) -> Metablock:
        return self.root.cli

    @property
    def url(self) -> str:
        return f"{self.cli.url}/{self.name}"


class CrudComponent(HttpComponent):
    Entity = MetablockEntity
    __slots__ = ("root", "name")

    def __init__(
        self, root: Metablock | CrudComponent | MetablockEntity, name: str = ""
    ) -> None:
        self.root = root
        self.name = name or self.__class__.__name__.lower()

    def __repr__(self) -> str:
        return self.url

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def cli(self) -> Metablock:
        return self.root.cli

    @property
    def url(self) -> str:
        return f"{self.cli.url}/{self.name}"

    @property
    def parent_url(self) -> str:
        return f"{self.root.url}/{self.name}"

    @property
    def is_entity(self) -> bool:
        return isinstance(self.root, MetablockEntity)

    async def execute(self, url: str | URL, **kwargs: Any) -> Any:
        return await self.root.execute(url, **kwargs)

    async def paginate(self, **params: Any) -> AsyncIterator[MetablockEntity]:
        url = self.list_create_url()
        next_ = url
        url_params = as_params(**params)
        while next_:
            if not next_.startswith(url):
                next_ = f'{url}?{next_.split("?")[1]}'
            data = await self.execute(next_, params=url_params)
            next_ = data.get("next")
            for d in data["data"]:
                yield self.wrap(d)

    async def get_list(self, **kwargs: Any) -> list[MetablockEntity]:
        url = self.list_create_url()
        kwargs.setdefault("wrap", self.wrap_list)
        return cast(
            list[MetablockEntity],
            await self.execute(url, **kwargs),
        )

    async def get_full_list(self, **kwargs: Any) -> list[MetablockEntity]:
        return [d async for d in self.paginate(**kwargs)]

    async def get(self, id_: str, **kwargs: Any) -> MetablockEntity:  # type: ignore
        url = f"{self.url}/{id_}"
        kwargs.setdefault("wrap", self.wrap)
        return cast(MetablockEntity, await self.execute(url, **kwargs))

    async def has(self, id_: str, **kwargs: Any) -> bool:
        url = f"{self.url}/{id_}"
        kwargs.update(wrap=self.head)
        return cast(bool, await self.get(url, callback=self.head))

    async def create(
        self, callback: Callback | None = None, **params: Any
    ) -> MetablockEntity:
        url = self.list_create_url()
        return cast(
            MetablockEntity,
            await self.post(url, json=params, callback=callback, wrap=self.wrap),
        )

    async def update(
        self, id_name: str, callback: Callback | None = None, **params: Any
    ) -> MetablockEntity:
        return cast(
            MetablockEntity,
            await self.patch(
                self.update_url(id_name), json=params, callback=callback, wrap=self.wrap
            ),
        )

    async def upsert(
        self, callback: Callback | None = None, **params: Any
    ) -> MetablockEntity:
        return cast(
            MetablockEntity,
            await self.put(self.url, json=params, callback=callback, wrap=self.wrap),
        )

    async def delete(  # type: ignore
        self,
        id_name: str,
        **kwargs: Any,
    ) -> Any:
        return await self.cli.delete(self.delete_url(id_name), **kwargs)

    async def delete_all(self) -> int:
        n = 0
        async for entity in self.paginate():
            await self.delete(entity.id)
            n += 1
        return n

    async def head(self, response: ClientResponse) -> bool:
        if response.status == 404:
            return False
        elif response.status == 200:
            return True
        else:  # pragma: no cover
            raise MetablockResponseError(response)

    def wrap(self, data: dict) -> MetablockEntity:
        return self.Entity(self, data)

    def entity(self, **data: Any) -> MetablockEntity:
        return self.Entity(self, data)

    def wrap_list(self, data: list[dict]) -> list[MetablockEntity]:
        return [self.wrap(d) for d in data]

    def list_create_url(self) -> str:
        return self.parent_url if self.is_entity else self.url

    def update_url(self, id_name: str) -> str:
        return f"{self.url}/{id_name}"

    def delete_url(self, id_name: str) -> str:
        return f"{self.url}/{id_name}"



from .client import Metablock
from .components import MetablockEntity, MetablockError, MetablockResponseError
from .extensions import Extension, Plugin
from .orgs import Org
from .spaces import Block, Service, Space, SpaceExtension
from .user import User

__version__ = "0.6.1"

__all__ = [
    "Metablock",
    "MetablockError",
    "MetablockResponseError",
    "MetablockEntity",
    "Space",
    "Service",
    "Block",
    "Extension",
    "Plugin",
    "SpaceExtension",
    "Org",
    "User",
]



from __future__ import annotations

import logging
import os
import sys
from typing import Any

from aiohttp import ClientSession
from yarl import URL

from .components import Callback, HttpComponent, MetablockResponseError
from .extensions import Extensions, Plugins
from .orgs import Orgs
from .spaces import Blocks, Domains, Space, Spaces
from .user import User

DEFAULT_USER_AGENT = f"Python/{'.'.join(map(str, sys.version_info[:2]))} metablock"

logger = logging.getLogger("metablock.client")


class Metablock(HttpComponent):
    """Metablock client"""

    url: str = os.environ.get("METABLOCK_URL", "https://api.metablock.io/v1")
    auth_key: str = os.getenv("METABLOCK_API_TOKEN", "")

    def __init__(
        self,
        url: str | None = None,
        auth_key: str = "",
        auth_key_name: str = "x-metablock-api-key",
        session: ClientSession | None = None,
        user_agent: str = DEFAULT_USER_AGENT,
    ) -> None:
        self.url: str = url if url is not None else self.url
        self.auth_key: str = auth_key or self.auth_key
        self.auth_key_name = auth_key_name
        self.session = session
        self.default_headers: dict[str, str] = {
            "user-agent": user_agent,
            "accept": "application/json",
        }
        self.spaces: Spaces = Spaces(self)
        self.domains = Domains(self)
        self.orgs: Orgs = Orgs(self)
        self.blocks: Blocks = Blocks(self, "services")
        self.services = self.blocks
        self.plugins: Plugins = Plugins(self)
        self.extensions: Extensions = Extensions(self)

    def __repr__(self) -> str:
        return self.url

    __str__ = __repr__

    @property
    def cli(self) -> Metablock:
        return self

    async def close(self) -> None:
        if self.session:
            await self.session.close()

    async def __aenter__(self) -> Metablock:
        return self

    async def __aexit__(self, exc_type: type, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    async def spec(self) -> dict:
        return await self.execute(f"{self.url}/spec")

    async def execute(
        self,
        url: str | URL,
        method: str = "",
        headers: dict[str, str] | None = None,
        callback: Callback | None = None,
        wrap: Any = None,
        **kw: Any,
    ) -> Any:
        if not self.session:
            self.session = ClientSession()
        method = method or "GET"
        headers_ = self.get_default_headers()
        headers_.update(headers or ())
        response = await self.session.request(method, url, headers=headers_, **kw)
        if callback:
            return await callback(response)
        if response.status == 204:
            return True
        if response.status >= 400:
            try:
                data = await response.json()
            except Exception:
                data = await response.text()
            raise MetablockResponseError(response, data)
        response.raise_for_status()
        data = await response.json()
        return wrap(data) if wrap else data

    async def get_user(self, **kw: Any) -> User:
        kw.setdefault("wrap", self._user)
        return await self.get(f"{self.url}/user", **kw)

    async def get_space(self, **kw: Any) -> Space:
        kw.setdefault("wrap", self._space)
        return await self.get(f"{self.url}/space", **kw)

    async def update_user(self, **kw: Any) -> User:
        kw.setdefault("wrap", self._user)
        return await self.patch(f"{self.url}/user", **kw)

    async def delete_user(self, **kw: Any) -> None:
        return await self.delete(f"{self.url}/user", **kw)

    def get_default_headers(self) -> dict[str, str]:
        headers = self.default_headers.copy()
        if self.auth_key:
            headers[self.auth_key_name] = self.auth_key
        return headers

    def _user(self, data: dict) -> User:
        return User(self, data)

    def _space(self, data: dict) -> Space:
        return Space(self.spaces, data)
